/*
 * milena_recogni.c - Milena TTS system (phraser)
 * Copyright (C) Bohdan R. Rau 2008 <ethanak@polip.com>
 * 
 * Milena is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * Milena is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Milena.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

/*

blank - cig spacji
\blank - opcjonalny cig spacji

\znak - opcjonalny znak

formaty:
	d - liczba
	r - liczba rzymska
	c - wybr
	e - wyraenie tekstowe

przykad:

recognize dmy {d:d1-31}.\ {m:r1-12}.\ {y:d}
recognize dmy {d:d1-31} {m:r1-12} {y:d}
recognize dmy {d:d1-31} {m:cmond} {y:d}
recognize dmyr {d:d1-31}.\ {m:r1-12}.\ {y:d} r.
recognize dmyr {d:d1-31} {m:r1-12} {y:d} r.
recognize wieczny {w:r1-30}-{f:ewieczny}	
expr wieczny wieczn(ymi|ym|ych|ego|emu|y|a||ej|e|i)

recognize ntego {n:d}-go

recognize dayoweek {w:cdow}
choice dow 1 pon
choice dow 1 poniedziaek
choiceval dow 1 poniedziaek

sayas:
	dXX - liczba we wbudowanym formacie
	
przykad:

sayas dmyr	{d:dmc} {m:dmt} {y:dmc} roku
sayas dmy	{d:dmc} {m:dmt} {y:dmc}
sayas wieczny	{w:dap}{f}
sayas ntego	{n:dmc}
sayas daoweek	{w}
*/


struct recog_format {
	struct recog_format *next;
	int flags;
	char *name;
	char *str;
};

static char *milena_mons[]={"","stycznia","lutego","marca","kwietnia",
	"maja","czerwca","lipca","sierpnia","wrzenia","padziernika",
	"listopada","grudnia",NULL};


/*
speaker_dit:
	mm - drugi
	fm - druga
	nm - drugie
	mc - drugiemu
	fd - drugiej
	md - drugiego
	fb - drug
	mn - drugim
	ap - drugo-
	mx - dwa
	fx - dwie
	f0 - zero dwie
	ac - dwu-
	mt - miesica
*/

#define MILDIT_MM 0
#define MILDIT_FM 1
#define MILDIT_NM 2
#define MILDIT_MC 3
#define MILDIT_FD 4
#define MILDIT_MD 5
#define MILDIT_FB 6
#define MILDIT_MN 7
#define MILDIT_AP 8
#define MILDIT_MX 9
#define MILDIT_FX 10
#define MILDIT_F0 11
#define MILDIT_AC 12
#define MILDIT_MT 13

#define MILRGT_EXPR 2
#define MILRGT_CHOICE 1

static char *dec_fin[]={
	"y","a","e","emu","ej","ego","","ym","o"};
	
static char *jed_por[]={
	"zerow","pierwsz","drugi","trzeci","czwart",
	"pit","szst","sidm","sm","dziewit",
	"dziesit","jedenast","dwunast","trzynast","czternast",
	"pitnast","szesnast","siedemnast","osiemnast","dziewitnast"};
static char *dec_por[]={
	NULL,NULL,"dwudziest","trzydziest","czterdziest",
	"pidziesit","szedziesit","siedemdziesit",
	"osiemdziesit","dziewidziesit"};
static char *dec_pac[]={
	NULL,NULL,"dwudziesto","trzydziesto","czterdziesto",
	"pidziesicio","szedziesicio","siedemdziesicio",
	"osiemdziesicio","dziewidziesicio"};
	
static char *hun_por[]={
	NULL,NULL,"dwu","trzech","czterech","pi","sze",
		"siedem","osiem","dziewi"};
static char *mil_por[]={
	NULL,"jedno","dwu","trzy","cztero","picio","szecio",
	"siedmio","omio","dziewicio","dziesicio"};
static char *hun_pac[]={NULL,"stu","dwustu","trzystu",
	"czterystu","piset","szeset","siedemset","osiemset","dziewiset"};
static char *hun_bac[]={NULL,"stu","dwustu","trzystu",
	"czterystu","piciuset","szeciuset","siedmiuset","omiuset","dziewiciuset"};

	
static int milena_speak_digit(int n,int mode,char *outbuf,int buflen,int pos)
{
	
	int m;
	if (n>999999) return milena_integer(n,outbuf,buflen,pos);
	if (mode == MILDIT_MT) {
		if (n>=1 && n<=12) pushstr(milena_mons[n]);
		return pos;
	}
	if (mode < MILDIT_MX) {
		if (!n) {
			pushstr(jed_por[0]);
			pushstr(dec_fin[mode]);
			return pos;
		}
		m=n%100;
		if (m) {
			if (n>100) {
				pos=milena_integer(n-m,outbuf,buflen,pos);
				pushbuf(' ');
			}
			if (m>=20) {
				pushstr(dec_por[m/10]);
				pushstr(dec_fin[mode]);
				if (m%10) pushbuf(' ');
				m%=10;
			}
			if (m==2 && mode == MILDIT_FM) {
				pushstr("druga");
			}
			else if (m==2 && mode == MILDIT_FB) {
				pushstr("drug");
			}
			else if (m==2 && mode == MILDIT_AP) {
				pushstr("drugo");
			}
			else if (m) {
				char *c=dec_fin[mode];
				pushstr(jed_por[m]);
				if ((m==2 || m==3) && *c=='y') c++;
				if (*c) pushstr(c);
			}
			return pos;
		}
		m=n%1000;
		if (m) {
			if (n>1000) {
				pos=milena_integer(n-m,outbuf,buflen,pos);
				pushbuf(' ');
			}
			m/=100;
			if (m>1) pushstr(hun_por[m]);
			pushstr("setn");
			pushstr(dec_fin[mode]);
			return pos;
		}
		if (n==1000) {
			if (mode == MILDIT_AP) {
				pushstr("tysic");
			}
			else {
				pushstr("tysiczn");
				pushstr(dec_fin[mode]);
			}
			return pos;
		}
		m=n/1000;
		pos=milena_speak_digit(m,MILDIT_AC,outbuf,buflen,pos);
		pushstr("tysiczn");
		pushstr(dec_fin[mode]);
		return pos;
	}
	
	
	switch(mode) {
		case MILDIT_F0: if (n<10) pushstr("zero ");
		case MILDIT_FX:
			m=n%100;
			if (m==12 || m%10!=2) return milena_integer(n,outbuf,buflen,pos);
			if (m!=2) {
				pos=milena_integer(n-2,outbuf,buflen,pos);
				pushbuf(' ');
			}
			pushstr("dwie");
			return pos;
		default: break;
	}
	if (mode == MILDIT_AC) {
		m=n%1000;
		if (!n) {
			pushstr("zero");
			return pos;
		}
		if (m) {
			if (n>1000) {
				pos=milena_integer(n-m,outbuf,buflen,pos);
				pushbuf(' ');
			}
			if (!m%100) {
				pushstr(hun_bac[m/100]);
				return pos;
			}
			if (m>=100) pushstr(hun_pac[m/100]);
			m%=100;
			if (!m) return pos;
			if (m<=10) {
				pushstr(mil_por[m]);
				return pos;
			}
			if (m<20) {
				pushstr(jed_por[m]);
				pushbuf('o');
				return pos;
			}
			pushstr(dec_pac[m/10]);
			m%=10;
			if (m) pushstr(mil_por[m]);
			return pos;
		}
		return milena_integer(n,outbuf,buflen,pos);
	}
	return milena_integer(n,outbuf,buflen,pos);
}

struct recog_result {
	short name;
	short mode;
	int value;	/* n dla int/choice, len dla expr */
	char *str;
};

struct recog_choice_item {
	struct recog_choice_item *next;
	int value;
	char *str;
};

struct recog_choice {
	struct recog_choice *next;
	char *name;
	struct recog_choice_item *items;
	struct recog_choice_item *values;
};

#define RECFLAG_EOS 1
#define RECFLAG_ATEND 2

struct recog_expression {
	struct recog_expression *next;
	char *name;
	char *str;
};

struct recog_sayas {
	struct recog_sayas *next;
	char *name;
	char *format;
};

static int milena_sayas(struct milena *cfg,char *format,struct recog_result *res,int nres,char *outbuf,int buflen,int pos)
{
	int i;
	static char *mr="mmfmnmmcfdmdfbmnapmxfxf0acmt";
	while (*format) {
		if (isspace(*format)) {
			pushbuf(' ');
			format++;
			while (*format && isspace(*format)) format++;
			continue;
		}
		if (*format!='{') {
			pushbuf(*format);
			format++;
			continue;
		}
		format++;
		for (i=0;i<nres;i++) if (res[i].name==*format) break;
		if (i>=nres) {
			format=strchr(format,'}');
			if (!format) return pos;
			format++;
			continue;
		}
		format++;
		if (*format==':') format++;
		if (res[i].mode == MILRGT_EXPR) {
			char *c=res[i].str;
			int j;
			for (j=0;j<res[i].value;j++) {
				char d=lci[c[j] & 255];
				if (!d) d=c[j];
				pushbuf(d);
			}
		}
		else if (res[i].mode == MILRGT_CHOICE && *format!='d' && *format!='u') {
			char *c=res[i].str;
			for (;*c;c++) {
				char d=lci[(*c) & 255];
				if (!d) d=*c;
				pushbuf(d);
			}
		}
		else if (*format=='u') {
			char *c;
			format++;
			c=milena_get_unit(cfg,&format,what_dit(res[i].value));
			if (c) {
				pushstr(c);
			}
			else {
				pushstr(format);
			}
			
			
		}
		else {
			int j,ftyp=MILDIT_MX;
			if (*format=='d') {
				format++;
				for (j=0;mr[2*j];j++) if (*format==mr[2*j] && format[1]==mr[2*j+1]) {
					ftyp=j;
					break;
				}
			}
			pos=milena_speak_digit(
					res[i].value,ftyp,outbuf,buflen,pos);
				
		}
		format=strchr(format,'}');
		if (!format) break;
		format++;
	}
	return pos;	
}

/* stub */
static int milena_roman(char **str)
{
	char *c=*str;
	int v=0;
	if (*c=='X') {
		c++;
		v=10;
		if (*c=='X') {c++;v+=10;}
		if (*c=='X') {c++;v+=10;}
	}
	if (*c=='V') {
		c++;
		v+=5;
		if (*c=='I') {c++;v++;}
		if (*c=='I') {c++;v++;}
		if (*c=='I') {c++;v++;}
		*str=c;
		return v;
	}
	if (*c=='I') {
		c++;
		if (*c=='V') {
			c++;
			v+=4;
			*str=c;
			return v;
		}
		if (*c=='X') {
			c++;
			v+=9;
			*str=c;
			return v;
		}
		v++;
		if (*c=='I') {c++;v++;}
		if (*c=='I') {c++;v++;}
	}
	if (v) *str=c;
	return v;
}

static char *milena_recognize_format(struct milena *cfg,char *str,
	struct recog_format *mft,struct recog_result *res,int *nres)
{
	char *s=mft->str;
	char *c,d,fbuf[32];
	int ndit,minval,maxval,mode,value;
	struct recog_expression *expr;
	struct recog_choice *choice;
	struct recog_choice_item *chitem,*chival;

	int compare_char(char pchar,char schar)
	{
		if (pchar == schar) return 1;
		if (uci[pchar & 255]) return 0;
		schar=lci[schar & 255];
		if (pchar == schar) return 1;
		return 0;
	}
	
	char *compare_expr(char *patr,char *str)
	{
		int i;
		while (*patr) {
			if (isspace(*patr)) {
				patr++;
				if (!*str || !isspace(*str)) return NULL;
				while (*str && isspace(*str)) str++;
				continue;
			}
			if (*patr=='_') {
				patr++;
				while (*str && isspace(*str)) str++;
				continue;
			}
			if (*patr!='(') {
				if (!compare_char(*patr++,*str++)) return NULL;
				continue;
			}
			patr++;
			for (;;) {
				for (i=0;patr[i] && patr[i]!='|' && patr[i]!=')';i++) {
					if (!compare_char(patr[i],str[i])) break;
				}
				if (!patr[i] || patr[i]=='|' || patr[i]==')') {
					if (!good_char(cfg,str[i])) {
						str+=i;
						break;
					}
				}
				while (*patr && *patr!='|' && *patr!=')') patr++;
				if (*patr++!='|') return NULL;
			}
			patr=strchr(patr,')');
			if (!patr) return str;
			patr++;
		}
		return str;
	}
	
	
	*nres=0;
	while (*s) {
		if (*s=='\\') {
			s++;
			if (!*s) break;
			if (isspace(*s)) {
				while (*str && isspace(*str)) str++;
				while (*s && isspace(*s)) s++;
				continue;
			}
			d=lci[(*str) & 255];
			if (!d) d=*str;
			if (d==*s) str++;
			
			s++;
			continue;
		}
		if (*s!='{') {
			if (isspace(*s)) {
				if (!*str || !isspace(*str)) return NULL;
				while (*str && isspace(*str)) str++;
				while(*s && isspace(*s)) s++;
				continue;
			}
			d=lci[(*str) & 255];
			if (!d) d=*str;
			if (d!=*s) return NULL;
			s++;str++;
			continue;
		}
		s++;
		if (!*s) return NULL;
		res[(*nres)].name=*s++;
		ndit=0;
		if (*s==':') s++;
		if (*s && isdigit(*s)) ndit=(*s++)-'0';
		mode=*s++;
		if (!mode || !strchr("drec",mode)) return NULL;
		if (mode=='r' || mode=='d') {
			minval=0;
			maxval=999999;
			if (isdigit(*s)) {
				minval=strtol(s,&s,10);
			}
			if (*s=='-' && isdigit(s[1])) {
				maxval=strtol(s+1,&s,10);
			}
			if (*s++!='}') return NULL;
			if (mode=='r') {
				value=milena_roman(&str);
				if (!value || value<minval || value>maxval) return NULL;
				res[(*nres)].value=value;
				res[(*nres)++].mode=0;
				continue;
			}
			if (!isdigit(*str)) return NULL;
			value=strtol(str,&c,10);
			if (value<minval || value>maxval) return NULL;
			if (ndit) {
				if (str+ndit != c) return NULL;
			}
			res[(*nres)].value=value;
			res[(*nres)++].mode=0;
			str=c;
			continue;
		}
		for (c=fbuf;*s && *s!='}';) *c++=*s++;
		if (*s++!='}') return NULL;
		*c=0;
		if (!*fbuf) return NULL;
		if (mode=='e') {
			for (expr=cfg->recexpr;expr;expr=expr->next) {
				if (strcmp(expr->name,fbuf)) continue;
				c=compare_expr(expr->str,str);
				if (c) break;
			}
			if (!expr) return NULL;
			res[(*nres)].value=c-str;
			res[(*nres)].str=str;
			str=c;
			res[(*nres)++].mode=MILRGT_EXPR;
			continue;
					
		}
		for (choice=cfg->choices;choice;choice=choice->next) if (!strcmp(choice->name,fbuf)) break;
		if (!choice) return NULL;
		for (chitem=choice->items;chitem;chitem=chitem->next) {
			c=compare_expr(chitem->str,str);
			if (c) break;
		}
		if (!chitem) return NULL;
		str=c;
		for (chival=choice->values;chival;chival=chival->next) if (chival->value==chitem->value) break;
		res[(*nres)].value=chitem->value;
		if (chival) {
			res[(*nres)].str=chival->str;
			res[(*nres)++].mode=MILRGT_CHOICE;
		}
		else {
			res[(*nres)++].mode=0;
		}
	}
	if (*str) {
		if (lci[(*str) & 255]) return NULL;
	}
	if (mft->flags & RECFLAG_ATEND) {
		if (!phraser_at_end(cfg,str)) return NULL;
	}
	if (mft->flags & RECFLAG_EOS) {
		if (*(str-1)=='.') {
			c=str;
			c++;
			while (*c && isspace(*c)) c++;
			if (*c && uci[(*c) & 255]) str--;
		}
	}
	return str;
}

static int milena_recognize(struct milena *cfg,char **str,char *outbuf,int buflen,int pos)
{
	char *c;
	struct recog_result res[16];
	int nres;
	struct recog_format *mft;
	struct recog_sayas *sya;
	for (mft=cfg->recs;mft;mft=mft->next) {
		c=milena_recognize_format(cfg,*str,mft,res,&nres);
		if (c) {
			for (sya=cfg->sayas;sya;sya=sya->next) if (!strcmp(sya->name,mft->name)) break;
			if (!sya) continue;
			if (pos) pushbuf(' ');
			pos=milena_sayas(cfg,sya->format,res,nres,outbuf,buflen,pos);
			*str=c;
			return pos;
		}
	}
	return -1;
}

static int milena_SkipRecognizedPart(struct milena *cfg,char **str)
{
	char *c;
	struct recog_result res[16];
	int nres;
	struct recog_format *mft;
	struct recog_sayas *sya;
	for (mft=cfg->recs;mft;mft=mft->next) {
		c=milena_recognize_format(cfg,*str,mft,res,&nres);
		if (c) {
			for (sya=cfg->sayas;sya;sya=sya->next) if (!strcmp(sya->name,mft->name)) break;
			if (!sya) continue;
			*str=c;
			return 1;
		}
	}
	return 0;
}


static void milena_init_recg(struct milena *cfg)
{
	int i;
	struct recog_choice *rc;
	struct recog_choice_item *rci;
	if (cfg->choices) return;
	rc=qalloc(sizeof(*rc));
	rc->next=NULL;
	rc->name="mond";
	cfg->choices=rc;
	rc->items=NULL;
	rc->values=NULL;
	for (i=1;i<=12;i++) {
		rci=qalloc(sizeof(*rci));
		rci->next=rc->items;
		rc->items=rci;
		rci->value=i;
		rci->str=milena_mons[i];
		rci=qalloc(sizeof(*rci));
		rci->next=rc->values;
		rc->values=rci;
		rci->value=i;
		rci->str=milena_mons[i];
	}
	
}

static int milena_line_is_recog(char *cmd)
{
	static char *mrec[]={
		"recognize","sayas","expr","choice","choiceval"};
	int i;
	for (i=0;i<5;i++) if (!strcmp(cmd,mrec[i])) return 1;
	return 0;
}

static void milena_recog_line(struct milena *cfg,char *cmd,char *str)
{
	char *rest=str;
	while (*rest && !isspace(*rest)) rest++;
	if (!*rest) gerror(cfg,"Blad skladni");
	*rest++=0;
	while (*rest && isspace(*rest)) rest++;
	if (!strcmp(cmd,"sayas")) {
		struct recog_sayas *sa;
		sa=qalloc(sizeof(*sa));
		sa->next=cfg->sayas;
		cfg->sayas=sa;
		sa->name=strdup(str);
		sa->format=strdup(rest);
		return;
	}
	if (!strcmp(cmd,"recognize")) {
		struct recog_format *ra;
		int flags=0;
		int atend=0;
		for (;*str;str++) {
			if (*str=='.') {
				flags |= RECFLAG_EOS;
				continue;
			}
			if (*str=='$') {
				flags |= RECFLAG_ATEND;
				continue;
			}
			if (*str=='+') {
				atend=1;
				continue;
			}
			break;
		}
		if (!*str) gerror(cfg,"Blad skladni");
		ra=qalloc(sizeof(*ra));
		if (atend) {
			struct recog_format **rra;
			for (rra=&cfg->recs;*rra;rra=&(*rra)->next);
			*rra=ra;
			ra->next=NULL;
		}
		else {
			ra->next=cfg->recs;
			cfg->recs=ra;
		}
		ra->flags = flags;
		ra->name=strdup(str);
		ra->str=strdup(rest);
		return;
	}
	if (!strcmp(cmd,"expr")) {
		struct recog_expression *expr;
		expr=malloc(sizeof(*expr));
		expr->next=cfg->recexpr;
		cfg->recexpr=expr;
		expr->name=strdup(str);
		sort_wordpar(rest);
		expr->str=strdup(rest);
		return;
	}
	if (!strcmp(cmd,"choice") || !strcmp(cmd,"choiceval")) {
		struct recog_choice *rc;
		struct recog_choice_item *rci;
		int val;
		if (!isdigit(*rest)) {
			gerror(cfg,"Oczekiwano liczby");
		}
		val=strtol(rest,&rest,10);
		while (*rest && isspace(*rest)) rest++;
		if (!*rest) {
			gerror(cfg,"Pusty choice");
		}
		for (rc=cfg->choices;rc;rc=rc->next) if (!strcmp(rc->name,str)) break;
		if (!rc) {
			rc=qalloc(sizeof(*rc));
			rc->next=cfg->choices;
			cfg->choices=rc;
			rc->items=NULL;
			rc->values=NULL;
			rc->name=strdup(str);
		}
		rci=qalloc(sizeof(*rci));
		rci->value=val;
		rci->str=strdup(rest);
		if (!strcmp(cmd,"choice")) {
			rci->next=rc->items;
			rc->items=rci;
		}
		else {
			rci->next=rc->values;
			rc->values=rci;
		}
		return;
	}
	
}
